package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyResourceNotFoundException;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.security.RootPermissions;
import ch.unige.solidify.specification.JoinedTableInValues;
import ch.unige.solidify.specification.SolidifySpecification;

@RootPermissions
public abstract class AssociationController<T extends Resource<T>, V extends Resource<V>> extends SubResourceController<T, V> {

  @GetMapping
  public HttpEntity<Collection<V>> list(@PathVariable String parentid, @ModelAttribute V filterItem, Pageable pageable) {
    SolidifySpecification<V> specification = this.subResourceService.getSpecification(filterItem);

    // add inner join to parent
    specification.addJoinTableInValues(new JoinedTableInValues(this.getParentFieldName(), SolidifyConstants.DB_RES_ID, Arrays.asList(parentid)));

    return this.getSubResourcesCollectionAsResponseEntity(parentid, specification, pageable);
  }

  @GetMapping(SolidifyConstants.URL_ID)
  public HttpEntity<V> get(@PathVariable String parentid, @PathVariable String id) {
    V subResource = this.findSubResource(parentid, id);
    this.addLinks(parentid, subResource);
    return new ResponseEntity<>(subResource, HttpStatus.OK);
  }

  @PostMapping
  public HttpEntity<List<V>> create(@PathVariable String parentid, @RequestBody String[] ids) {
    boolean noError = true;
    final T parentItem = this.resourceService.findOne(parentid);

    final List<V> list = new ArrayList<>();
    for (final String id : ids) {
      final V childItem = this.subResourceService.findOne(id);
      if (childItem == null) {
        return new ResponseEntity<>(HttpStatus.NOT_FOUND);
      }
      noError &= this.addChildOnParent(parentItem, childItem);
      list.add(childItem);
    }

    // Save entity if no error
    if (noError) {
      this.resourceService.save(parentItem);
      return new ResponseEntity<>(list, HttpStatus.CREATED);
    }
    return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
  }

  @PatchMapping(SolidifyConstants.URL_ID)
  public HttpEntity<V> update(@PathVariable String parentid, @PathVariable String id, @RequestBody V v2) {
    return new ResponseEntity<>(HttpStatus.NOT_IMPLEMENTED);
  }

  @DeleteMapping(SolidifyConstants.URL_ID)
  public HttpStatus delete(@PathVariable String parentid, @PathVariable String id) {
    return this.deleteList(parentid, new String[] { id });
  }

  @DeleteMapping
  public HttpStatus deleteList(@PathVariable String parentid, @RequestBody String[] ids) {
    boolean noError = true;
    final T parentItem = this.resourceService.findOne(parentid);

    for (final String id : ids) {
      V subResource = this.findSubResource(parentid, id);
      noError &= this.removeChildFromParent(parentItem, subResource);
    }

    if (noError) {
      this.resourceService.save(parentItem);
      return HttpStatus.OK;
    }
    throw new SolidifyHttpErrorException(HttpStatus.GONE, "");
  }

  @Override
  protected void addLinks(String parentid, V v) {
    v.removeLinks();
    v.addLinks(linkTo(methodOn(this.getClass()).list(parentid, null, null)), false, false);
  }

  private V findSubResource(String parentId, String id) {
    V subResource = this.subResourceService.findOne(id);
    SolidifySpecification<V> specification = this.subResourceService.getSpecification(subResource);

    // generate a where clause with resId = id
    specification.setFilterOnRootResId(true);
    // add inner join to parent
    specification.addJoinTableInValues(new JoinedTableInValues(this.getParentFieldName(), SolidifyConstants.DB_RES_ID, Arrays.asList(parentId)));

    final Page<V> list = this.subResourceService.findAll(specification, Pageable.unpaged());
    if (list.getTotalElements() == 1) {
      return list.getContent().get(0);
    } else {
      throw new SolidifyResourceNotFoundException("Child item with id " + id + " not found");
    }
  }

  protected abstract String getParentFieldName();

  protected abstract boolean addChildOnParent(T parentResource, V childResource);

  protected abstract boolean removeChildFromParent(T parentResource, V childResource);
}
