package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.exception.SolidifyUndeletableException;
import ch.unige.solidify.exception.SolidifyUnmodifiableException;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.security.RootPermissions;

@RootPermissions
public abstract class ResourceController<T extends Resource<T>> extends ResourceReadOnlyController<T> implements ControllerWithHateoasHome {

  @PostMapping
  public HttpEntity<T> create(@RequestBody T t) {

    if (t.getResId() != null && this.itemService.existsById(t.getResId())) {
      throw new SolidifyRuntimeException(this.messageService.get("validation.resource.alreadyexist",
              new Object[] { t.getResId() }));
    }

    final T item = this.itemService.save(t);
    this.addLinks(item);
    return new ResponseEntity<>(item, HttpStatus.CREATED);
  }

  @DeleteMapping(SolidifyConstants.URL_ID)
  public HttpStatus delete(@PathVariable String id) {
    final T t = this.itemService.findOne(id);

    if (!t.isDeletable()) {
      throw new SolidifyUndeletableException(
              this.messageService.get("validation.resource.undeletable",
                      new Object[] { t.getClass().getSimpleName() + ": " + id }));
    }

    this.itemService.delete(id);
    return HttpStatus.OK;
  }

  @DeleteMapping
  public HttpStatus deleteList(@RequestBody String[] ids) {
    // Remove each item
    for (final String id : ids) {
      final T t = this.itemService.findOne(id);
      if (!t.isDeletable()) {
        throw new SolidifyUndeletableException(
                this.messageService.get("validation.resource.undeletable",
                        new Object[] { t.getClass().getSimpleName() + ": " + id }));
      }
      this.itemService.delete(id);
    }
    return HttpStatus.OK;
  }

  @PatchMapping(SolidifyConstants.URL_ID)
  public HttpEntity<T> update(@PathVariable String id, @RequestBody T t2) {

    final T t1 = this.itemService.findOne(id);

    if (!t1.isModifiable()) {
      throw new SolidifyUnmodifiableException(
              this.messageService.get("validation.resource.unmodifiable",
                      new Object[] { t1.getClass().getSimpleName() + ": " + id }));
    }

    if (t1.update(t2)) {
      final T item = this.itemService.save(t1);
      this.addLinks(item);
      return new ResponseEntity<>(item, HttpStatus.OK);
    } else {
      return new ResponseEntity<>(t1, HttpStatus.OK);
    }
  }

  @Override
  protected void addLinks(T t) {
    t.removeLinks();
    t.addLinks(linkTo(this.getClass()), true, true);
  }

  @Override
  protected <W extends RepresentationModel<W>> void addOthersLinks(W w) {
    // Do nothing
    // Could be overridden by subclass
  }

}
