package ch.unige.solidify.service;

import static ch.unige.solidify.util.StringTool.isNullOrEmpty;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.BindingResult;

import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.validation.ValidationError;

/**
 * A service for composite resources
 *
 * @param <T> the type of the resource
 */
public abstract class CompositeResourceService<T extends Resource<T>> extends ResourceService<T> {

  private static final Logger log = LoggerFactory.getLogger(CompositeResourceService.class);

  @Autowired
  protected SolidifyProperties config;

  @Autowired
  protected MessageService messageService;

  /**
   * Use a LinkedRESTResourceChecker instead of calling directly validateLinkedRESTResources() method
   * in order to be able to disable validation during tests. (during tests
   * TestLinkedRESTResourceChecker is injected, thus bypassing linked resources validation)
   */
  @Autowired
  private LinkedResourceChecker<T> linkedResourceValidator;

  @Override
  public Page<T> findAll(Pageable pageable) {
    Page<T> page = super.findAll(pageable);
    this.completeEmbededResources(page);
    return page;
  }

  @Override
  public Page<T> findAll(Specification<T> spec, Pageable pageable) {
    Page<T> page = super.findAll(spec, pageable);
    this.completeEmbededResources(page);
    return page;
  }

  @Override
  public T findOne(String resId) {
    final T item = super.findOne(resId);
    this.completeEmbededResources(item);
    return item;
  }

  /**
   * Complete a page of resources by adding all its subresources
   *
   * @param items
   */
  protected void completeEmbededResources(Page<T> items) {
    for (final T item : items) {
      this.completeEmbededResources(item);
    }
  }

  /**
   * Complete a resource by adding all its subresources
   *
   * @param item the resource to be completed
   */
  protected void completeEmbededResources(T item) {
    if (item instanceof SubResourceContainer) {
      final List<Class<? extends Resource<?>>> embeddedResources = ((SubResourceContainer) item).getEmbeddedResourceTypes();
      for (final Class<? extends Resource> type : embeddedResources) {
        this.completeEmbededResources(item, type);
      }
    }
  }

  /**
   * Complete a resource by adding a subresource of a given type
   *
   * @param item            the resource to be completed
   * @param subResourceType the subresource type
   * @param <V>             The subresource type as a parameter type
   */
  protected <V extends Resource<V>> void completeEmbededResources(T item,
          Class<V> subResourceType) {

    if (item instanceof SubResourceContainer) {

      try {

        /*
         * Get sub resource Id from item property (i.e: getLanguageId)
         */
        String methodName = "get" + subResourceType.getSimpleName() + "Id";
        final Method getter = item.getClass().getMethod(methodName);
        final String subResourceId = (String) getter.invoke(item);

        if (!isNullOrEmpty(subResourceId)) {

          final Resource<?> child = this.getSubResource(subResourceId, subResourceType);

          if (child == null) {
            log.error(this.messageService.get("resources.embedded.notfound",
                    new Object[] { subResourceId, subResourceType.getName(), item.getResId(),
                            item.getClass().getName() }));
          }

          /*
           * Set the item property with fetched object
           */
          methodName = "set" + subResourceType.getSimpleName();
          final Method setter = item.getClass().getMethod(methodName, subResourceType);
          setter.invoke(item, child);
        }

      } catch (final Exception ex) {
        log.error("Unable to retrieve subresource " + subResourceType.getSimpleName() + " "
                + item.getResId());
      }
    } else {
      throw new IllegalArgumentException("item must be a SubResourceContainer");
    }
  }

  public <S extends Resource<S>> S getSubResource(String subResourceId, Class<S> subResourceType) {
    final RemoteResourceService<S> service = this.getResourceService(subResourceType);
    return service.findOneWithCache(subResourceId);
  }

  public <S extends Resource<S>, U extends SubResourceContainer> Collection<S> getSubResources(U parentItem, Class<S> subResourceType,
          Pageable pageable) {

    Page<String> subResourceIds = parentItem.getSubResourceIds(subResourceType, pageable);

    final List<S> items = new ArrayList<>();

    if (subResourceIds != null) {
      for (final String subResourceId : subResourceIds) {
        if (!isNullOrEmpty(subResourceId)) {
          final S item = this.getSubResource(subResourceId, subResourceType);
          if (item != null) {
            items.add(item);
          }
        }
      }
      Page page = new PageImpl(items, pageable, subResourceIds.getTotalElements());
      return new Collection<>(page, pageable);
    } else {
      return new Collection<>();
    }
  }

  protected <S extends Resource<S>> RemoteResourceService<S> getResourceService(Class<S> type) {
    throw new UnsupportedOperationException("Missing resource service for type " + type.getName());
  }

  @Override
  protected void validate(T item) {
    super.validate(item);
    final BindingResult errors = new BeanPropertyBindingResult(item, item.getClass().getName());
    this.linkedResourceValidator.validateLinkedResources(this, item, errors);
    if (errors.hasErrors()) {
      throw new SolidifyValidationException(new ValidationError(errors));
    }
  }
}
