package ch.unige.solidify.listener;

import javax.persistence.PostPersist;
import javax.persistence.PostRemove;
import javax.persistence.PostUpdate;

import org.springframework.context.ApplicationEventPublisher;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.message.CacheMessage;
import ch.unige.solidify.rest.JoinResource;
import ch.unige.solidify.rest.Resource;

public class CacheListener {

  @PostPersist
  /**
   * Publish a CacheMessage event that may be used to invalidate cache when a new relations between entities is created.
   * The message is sent only if the object is a JoinResource as there is no need to invalidate cache for new resources.
   */
  private void notifyPersist(Object object) {
    final ApplicationEventPublisher publisher = SolidifyEventPublisher.getPublisher();

    // publisher may be null when creating resources during initialization
    if (publisher != null && object instanceof JoinResource) {
      CacheMessage cacheMessage = ((JoinResource<?, ?>) object).getCacheMessage();
      if (cacheMessage != null) {
        publisher.publishEvent(cacheMessage);
      }
    }
  }

  @PostUpdate
  @PostRemove
  /**
   * Publish a CacheMessage event that may be used to invalidate cache when object is updated or deleted.
   * The message is sent if the object is a simple Resource or a JoinResource as it may worth to invalidate cache in both cases.
   */
  private void notifyUpdate(Object object) {
    final ApplicationEventPublisher publisher = SolidifyEventPublisher.getPublisher();

    // publisher may be null when creating resources during initialization
    if (publisher != null) {
      CacheMessage cacheMessage = null;
      if (object instanceof Resource) {
        cacheMessage = ((Resource<?>) object).getCacheMessage();
      } else if (object instanceof JoinResource) {
        cacheMessage = ((JoinResource<?, ?>) object).getCacheMessage();
      }
      if (cacheMessage != null) {
        publisher.publishEvent(cacheMessage);
      }
    }
  }
}
