package ch.unige.solidify.model;

import static org.springframework.util.Assert.hasText;

import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.validation.constraints.NotNull;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ResourceBase;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.rest.SolidifyModuleName;
import ch.unige.solidify.util.ReflectionTool;
import ch.unige.solidify.util.StringTool;

@Entity
public class RoleApplication extends ResourceNormalized<RoleApplication> {

  public static final RoleApplication ROOT = new RoleApplication("Root", 10);
  public static final RoleApplication TRUSTED_CLIENT = new RoleApplication("Trusted_client", 20);
  public static final RoleApplication ADMIN = new RoleApplication("Admin", 30);
  public static final RoleApplication USER = new RoleApplication("User", 40);

  public static final String ROOT_ID = "ROOT";
  public static final String ADMIN_ID = "ADMIN";
  public static final String TRUSTED_CLIENT_ID = "TRUSTED_CLIENT";
  public static final String USER_ID = "USER";

  private static final List<RoleApplication> roleApplicationList = Arrays.asList(ROOT, ADMIN, TRUSTED_CLIENT, USER);

  @NotNull
  @Column(unique = true)
  private String name;

  @NotNull
  @Column(unique = true)
  private int level;

  public RoleApplication() {
  }

  public RoleApplication(String name, int level) {
    hasText(name, "A role representation is required");
    this.setResId(name.toUpperCase());
    this.name = StringTool.toCamelCase(name);
    this.level = level;
  }

  public String getName() {
    return this.name;
  }

  public int getLevel() {
    return this.level;
  }

  public static List<RoleApplication> getRoleApplicationList() {
    return roleApplicationList;
  }

  public static RoleApplication getRoleFromName(String roleName) {
    if (roleName.equalsIgnoreCase(ADMIN.name)) {
      return ADMIN;
    }
    if (roleName.equalsIgnoreCase(ROOT.name)) {
      return ROOT;
    }
    if (roleName.equalsIgnoreCase(TRUSTED_CLIENT.name)) {
      return TRUSTED_CLIENT;
    }
    if (roleName.equalsIgnoreCase(USER.name)) {
      return USER;
    }
    throw new SolidifyRuntimeException("Unkown role : " + roleName);
  }

  @Override
  public void init() {
    // Do nothing
  }

  @Override
  public String managedBy() {
    return SolidifyModuleName.ADMIN;
  }

  public void setName(String name) {
    this.setProperty(ReflectionTool.getCurrentSetterVariableName(), name);
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    builder.append("Application Role [name=").append(this.name).append("]").append("[resId=")
            .append(this.getResId()).append("]");
    return builder.toString();
  }

  @Override
  public boolean update(ResourceBase item) {
    // Level cannot be updated
    if (((RoleApplication) item).getLevel() != this.getLevel()) {
      return false;
    } else {
      return super.update(item);
    }
  }

  @Override
  public boolean equals(final Object o) {
    if (this == o)
      return true;
    if (!(o instanceof RoleApplication))
      return false;
    final RoleApplication that = (RoleApplication) o;
    return this.getResId().equals(that.getResId());
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.name, this.level);
  }
}
