package ch.unige.solidify.test.util;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;

import java.util.List;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import ch.unige.solidify.util.ReflectionTool;

@ExtendWith(SpringExtension.class)
class ReflectionToolTest {

  private class Thing {
    private String name;
    private Integer age;

    public String getName() {
      return this.name;
    }

    public void setName(String name) {
      this.name = name;
    }

    public Integer getAge() {
      return this.age;
    }

    public void setAge(Integer age) {
      this.age = age;
    }
  }

  private enum InstrumentType {STRING, WIND, PERCUSSION}

  private class Instrument extends Thing {
    private InstrumentType type;

    public InstrumentType getType() {
      return this.type;
    }

    public void setType(InstrumentType type) {
      this.type = type;
    }
  }

  private class StringInstrument extends Instrument {
    public StringInstrument() {
      this.setType(InstrumentType.STRING);
    }
  }

  private class Guitar extends StringInstrument {
    private int strings;

    public int getStrings() {
      return this.strings;
    }

    public void setStrings(int strings) {
      this.strings = strings;
    }
  }

  private class PercussionInstrument extends Instrument {
    public PercussionInstrument() {
      this.setType(InstrumentType.PERCUSSION);
    }
  }

  private class Drums extends PercussionInstrument {
    private int toms;

    public int getToms() {
      return this.toms;
    }

    public void setToms(int toms) {
      this.toms = toms;
    }
  }

  private enum Size {S, M, L}

  private class TShirt extends Thing {
    private Size size;

    public Size getSize() {
      return this.size;
    }

    public void setSize(Size size) {
      this.size = size;
    }
  }

  @Test
  void getParentClassesTest() {
    Guitar guitar = new Guitar();
    List<Class> guitarParentClasses = ReflectionTool.getParentClasses(guitar);
    assertEquals(5, guitarParentClasses.size());
    assertEquals(Guitar.class, guitarParentClasses.get(0));
    assertEquals(StringInstrument.class, guitarParentClasses.get(1));
    assertEquals(Instrument.class, guitarParentClasses.get(2));
    assertEquals(Thing.class, guitarParentClasses.get(3));
    assertEquals(Object.class, guitarParentClasses.get(4));

    Drums drums = new Drums();
    List<Class> drumsParentClasses = ReflectionTool.getParentClasses(drums);
    assertEquals(5, drumsParentClasses.size());
    assertEquals(Drums.class, drumsParentClasses.get(0));
    assertEquals(PercussionInstrument.class, drumsParentClasses.get(1));
    assertEquals(Instrument.class, drumsParentClasses.get(2));
    assertEquals(Thing.class, drumsParentClasses.get(3));
    assertEquals(Object.class, drumsParentClasses.get(4));

    List<Class> newGuitarParentClasses = ReflectionTool.getParentClasses(guitar, Instrument.class);
    assertEquals(3, newGuitarParentClasses.size());
    assertEquals(Guitar.class, newGuitarParentClasses.get(0));
    assertEquals(StringInstrument.class, newGuitarParentClasses.get(1));
    assertEquals(Instrument.class, newGuitarParentClasses.get(2));
  }

  @Test
  void copyFieldsTest() {
    Guitar guitar = new Guitar();
    guitar.setName("guitar");
    guitar.setAge(12);
    assertEquals(InstrumentType.STRING, guitar.getType());

    // all properties are copied
    Drums drums = new Drums();
    drums.setToms(5);
    assertEquals(InstrumentType.PERCUSSION, drums.getType());
    ReflectionTool.copyFields(drums, guitar, Thing.class);
    assertEquals(5, drums.getToms());
    assertEquals(InstrumentType.STRING, drums.getType());
    assertEquals("guitar", drums.getName());
    assertEquals(12, drums.getAge().intValue());

    // name and age are not copied as they are Thing's members, but type is copied
    drums = new Drums();
    drums.setToms(5);
    assertEquals(InstrumentType.PERCUSSION, drums.getType());
    ReflectionTool.copyFields(drums, guitar, Instrument.class);
    assertEquals(5, drums.getToms());
    assertEquals(InstrumentType.STRING, drums.getType());
    assertNull(drums.getName());
    assertNull(drums.getAge());

    // all properties are copied
    TShirt tShirt = new TShirt();
    tShirt.setSize(Size.L);
    ReflectionTool.copyFields(tShirt, guitar, Thing.class);
    assertEquals(Size.L, tShirt.getSize());
    assertEquals("guitar", tShirt.getName());
    assertEquals(12, tShirt.getAge().intValue());

    // name and age are not copied as they are Thing's members
    tShirt = new TShirt();
    tShirt.setSize(Size.L);
    ReflectionTool.copyFields(tShirt, guitar, Instrument.class);
    assertEquals(Size.L, tShirt.getSize());
    assertNull(tShirt.getName());
    assertNull(tShirt.getAge());
  }
}
